'''
OFTests for functionality needed from the OLT.
'''
import logging
from oftest import config
import oftest.base_tests as base_tests
import oftest.packet as scapy

import ofp
import time

from oftest.testutils import *

onu_port = test_param_get("onu_port", 1)
olt_port = test_param_get("olt_port", 129)

def testPacketIn(self, match, parsed_pkt):

    delete_all_flows(self.controller)

    pkt = str(parsed_pkt)

    request = ofp.message.flow_add(
        table_id=test_param_get("table", 0),
        cookie=42,
        match=match,
        instructions=[
            ofp.instruction.apply_actions(
                actions=[
                    ofp.action.output(
                        port=ofp.OFPP_CONTROLLER,
                        max_len=ofp.OFPCML_NO_BUFFER)])],
        buffer_id=ofp.OFP_NO_BUFFER,
        priority=1000)

    logging.info("Inserting flow sending matching packets to controller")
    self.controller.message_send(request)
    do_barrier(self.controller)
	
    for of_port in config["port_map"]:
        logging.info("PacketInExact test, port %d", of_port)
        self.dataplane.send(of_port, pkt)
        verify_packet_in(self, pkt, of_port, ofp.OFPR_ACTION)
        verify_packets(self, pkt, [])

class EapolPacketIn(base_tests.SimpleDataPlane):
    
    """Verify packet-ins are sent for EAPOL packets """
    
    def runTest(self):
        logging.info("Running EAPOL Packet In test")

        match = ofp.match()
        match.oxm_list.append(ofp.oxm.eth_type(0x888e))
        # Use ethertype 0x888e and multicast destination MAC address
        pkt = simple_eth_packet(pktlen=60, eth_dst='01:00:5E:7F:FF:FF', eth_type=0x888e)
        
        testPacketIn(self, match, pkt)

class ARPPacketIn(base_tests.SimpleDataPlane):
    
    """Verify packet-ins are sent for ARP packets """
    
    def runTest(self):
        logging.info("Running ARP Packet In test")

        match = ofp.match()
        match.oxm_list.append(ofp.oxm.eth_type(0x0806))

        # Use ethertype 0x0806
        pkt = simple_eth_packet(eth_type=0x0806)
        
        testPacketIn(self, match, pkt)
        
class DHCPPacketIn(base_tests.SimpleDataPlane):
    
    """Verify packet-ins are sent for DHCP packets """
    
    def runTest(self):
        logging.info("Running DHCP Packet In test")

        pkt = simple_udp_packet(udp_sport=68, udp_dport=67)
        
        match = ofp.match()
        match.oxm_list.append(ofp.oxm.eth_type(0x0800))
        match.oxm_list.append(ofp.oxm.ip_proto(17))
        match.oxm_list.append(ofp.oxm.udp_src(68))
        match.oxm_list.append(ofp.oxm.udp_dst(67))
        
        testPacketIn(self, match, pkt)
        
        # Other UDP packets should not match the rule
        radiusPkt = simple_udp_packet(udp_sport=1812, udp_dport=1812)
        
        self.dataplane.send(1, str(radiusPkt))
        verify_no_packet_in(self, radiusPkt, 1)    
        
class IGMPPacketIn(base_tests.SimpleDataPlane):
    
    """Verify packet-ins are sent for IGMP packets """
    
    def runTest(self):
        logging.info("Running IGMP Packet In test")

        match = ofp.match()
        match.oxm_list.append(ofp.oxm.eth_type(0x800))
        match.oxm_list.append(ofp.oxm.ip_proto(2))
        
        pkt = scapy.Ether(dst='01:00:5E:7F:FF:FF', src='00:00:00:00:00:01')/ \
            scapy.IP(src='10.0.0.1', dst='10.0.0.2', ttl=60, tos=0, id=0, proto=2)

        pkt = pkt/("0" * (100 - len(pkt)))
        
        testPacketIn(self, match, pkt)

class TestMeter(base_tests.SimpleDataPlane):
       
    def runTest(self):
        logging.info("Running Meter tests")
        dropMeterBand = ofp.meter_band.drop(rate = 500)
        meter_mod = ofp.message.meter_mod(xid = 1, command = ofp.OFPMC_ADD, meter_id = 1, meters = [ dropMeterBand ])
        self.controller.message_send(meter_mod)
    
        time.sleep(1)

        verify_no_errors(self.controller)

        match = ofp.match()
        match.oxm_list.append(ofp.oxm.in_port(onu_port))
        match.oxm_list.append(ofp.oxm.vlan_vid(ofp.OFPVID_PRESENT | 201))

        request = ofp.message.flow_add(
            table_id=test_param_get("table", 0),
            cookie=43,
            match=match,
            instructions=[
                ofp.instruction.apply_actions(
                    actions=[
                        ofp.action.output(port=olt_port)]),
                    ofp.instruction.meter(meter_id = 1)    
                ],
            buffer_id=ofp.OFP_NO_BUFFER,
            priority=1000)

        self.controller.message_send(request)
        
        time.sleep(1)

        verify_no_errors(self.controller)
        
class TestDuplicateMeter(base_tests.SimpleDataPlane):

    def runTest(self):
        logging.info("Running Duplicate Meter Test")
        dropMeterBand = ofp.meter_band.drop(rate = 500)
        meter_mod = ofp.message.meter_mod(xid = 1, command = ofp.OFPMC_ADD, meter_id = 1, meters = [ dropMeterBand ])
        self.controller.message_send(meter_mod)
        self.controller.message_send(meter_mod)

        time.sleep(1)
    
        try:
            verify_no_errors(self.controller)
        except AssertionError as e:
            if (not e.message is "unexpected error type=12 code=1"):
                raise AssertionError("Incorrect error type: %s" % e.message)
            
 
class VlanTest(base_tests.SimpleDataPlane):
    
    """Verify the switch can push/pop a VLAN tag and forward out a port """
    
    def runTest(self):
        logging.info("Running push VLAN test")
        
        vlan_id = 200
        
        delete_all_flows(self.controller)
       
        #PUSH
        match = ofp.match()
        match.oxm_list.append(ofp.oxm.in_port(onu_port))
        match.oxm_list.append(ofp.oxm.vlan_vid_masked(value=ofp.OFPVID_PRESENT, value_mask=ofp.OFPVID_PRESENT))
        match.oxm_list.append(ofp.oxm.vlan_pcp(value = 0))
        
        request = ofp.message.flow_add(
            table_id=test_param_get("table", 0),
            cookie=42,
            match=match,
            instructions=[
                ofp.instruction.apply_actions(
                    actions=[
                        ofp.action.push_vlan(ethertype=0x8100),
                        ofp.action.set_field(ofp.oxm.vlan_vid(ofp.OFPVID_PRESENT | vlan_id)),
                        ofp.action.set_field(ofp.oxm.vlan_pcp(0)),
                        ofp.action.output(port=olt_port)])],
            buffer_id=ofp.OFP_NO_BUFFER,
            priority=1000)

        logging.info("Inserting flow tagging upstream")
        self.controller.message_send(request)

   	    #POP
        match = ofp.match()
        match.oxm_list.append(ofp.oxm.in_port(olt_port))
        match.oxm_list.append(ofp.oxm.vlan_vid(ofp.OFPVID_PRESENT | vlan_id))

        request = ofp.message.flow_add(
            table_id=test_param_get("table", 0),
            cookie=43,
            match=match,
            instructions=[
                ofp.instruction.apply_actions(
                    actions=[
                        ofp.action.pop_vlan(),
                        ofp.action.output(port=1)])],
            buffer_id=ofp.OFP_NO_BUFFER,
            priority=1000)

        logging.info("Inserting flow tagging downstream")
        self.controller.message_send(request)
        do_barrier(self.controller)
        time.sleep(5)
        
        inPkt = simple_udp_packet(dl_vlan_enable=True, vlan_vid=0, vlan_pcp=0)
        outPkt = simple_udp_packet(pktlen=100, dl_vlan_enable=True, 
                                   vlan_vid=vlan_id, vlan_pcp=0, dl_vlan_cfi=0)
        
        # Send untagged packet in the ONU port and expect tagged packet out the OLT port
        self.dataplane.send(onu_port, str(inPkt))
        verify_packet(self, outPkt, olt_port)
        
        # Send untagged packet in the OLT port and expect no packets to come out
        self.dataplane.send(olt_port, str(inPkt))
        verify_packets(self, outPkt, [])

        inPkt = simple_udp_packet(pktlen=104, dl_vlan_enable=True,
                                  vlan_vid=vlan_id, vlan_pcp=0, dl_vlan_cfi=0)
        outPkt = simple_udp_packet(pktlen=104, dl_vlan_enable=True, vlan_vid=0, vlan_pcp=0)

        # Send tagged packet in the OLT port and expect untagged packet out the OLT port
        self.dataplane.send(olt_port, str(inPkt))
        verify_packet(self, outPkt, onu_port)

        # Send tagged packet in the ONU port and expect no packets to come out
        self.dataplane.send(onu_port, str(inPkt))
        verify_packets(self, outPkt, [])
        
